@@ -40,7 +40,7 @@
 
 #define CERT_EXPIRY_TIME (60*60*24*30) /* 30 days */
 
-struct dtls_connection *dtls_ptr(struct stream_fd *sfd) {
+struct dtls_connection *dtls_ptr(stream_fd *sfd) {
 	if (!sfd)
 		return NULL;
 	struct packet_stream *ps = sfd->stream;
@@ -377,7 +377,7 @@ err:
 	return -1;
 }
 
-int dtls_init() {
+int dtls_init(void) {
 	int i;
 	char *p;
 
@@ -460,7 +460,7 @@ static unsigned int sha_512_func(unsigned char *o, X509 *x) {
 }
 
 
-struct dtls_cert *dtls_cert() {
+struct dtls_cert *dtls_cert(void) {
 	struct dtls_cert *ret;
 
 	rwlock_lock_r(&__dtls_cert_lock);
@@ -766,8 +766,8 @@ found:
 			crypto_init(&ps->selected_sfd->crypto, &client);
 	}
 	// it's possible that ps->selected_sfd is not from ps->sfds list (?)
-	for (GList *l = ps->sfds.head; l; l = l->next) {
-		struct stream_fd *sfd = l->data;
+	for (__auto_type l = ps->sfds.head; l; l = l->next) {
+		stream_fd *sfd = l->data;
 		if (d->active) /* we're the client */
 			crypto_init(&sfd->crypto, &server);
 		else /* we're the server */
@@ -789,7 +789,7 @@ error:
 }
 
 /* called with call locked in W or R with ps->in_lock held */
-int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
+int dtls(stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 	struct packet_stream *ps = sfd->stream;
 	int ret;
 	unsigned char buf[0x10000];
@@ -884,8 +884,8 @@ int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 		if (fsin) {
 			ilogs(srtp, LOG_DEBUG, "Sending DTLS packet");
 			socket_sendto(&sfd->socket, buf, ret, fsin);
-			atomic64_inc(&ps->stats_out.packets);
-			atomic64_add(&ps->stats_out.bytes, ret);
+			atomic64_inc_na(&ps->stats_out->packets);
+			atomic64_add_na(&ps->stats_out->bytes, ret);
 		}
 	}
 
@@ -908,8 +908,8 @@ void dtls_shutdown(struct packet_stream *ps) {
 		}
 		dtls_connection_cleanup(&ps->ice_dtls);
 	}
-	for (GList *l = ps->sfds.head; l; l = l->next) {
-		struct stream_fd *sfd = l->data;
+	for (__auto_type l = ps->sfds.head; l; l = l->next) {
+		stream_fd *sfd = l->data;
 
 		struct dtls_connection *d = &sfd->dtls;
 		if (!d->init)
