@@ -1,6 +1,8 @@
 #include "janus.h"
+
 #include <json-glib/json-glib.h>
 #include <stdbool.h>
+
 #include "websocket.h"
 #include "log.h"
 #include "main.h"
@@ -12,46 +14,65 @@
 #include "ice.h"
 #include "log_funcs.h"
 
+TYPED_GHASHTABLE(janus_handles_set, uint64_t, void, int64_hash, int64_eq, NULL, NULL)
+TYPED_DIRECT_FUNCS(websocket_conn_direct_hash, websocket_conn_direct_eq, struct websocket_conn)
+TYPED_GHASHTABLE(janus_websockets_ht, struct websocket_conn, struct websocket_conn,
+		websocket_conn_direct_hash, websocket_conn_direct_eq, NULL, NULL)
 
 struct janus_session { // "login" session
 	struct obj obj;
 	uint64_t id;
 	mutex_t lock;
 	time_t last_act;
-	GHashTable *websockets; // controlling transports, websocket_conn -> websocket_conn
-	GHashTable *handles; // handle ID -> 0x1. handle ID owned by janus_handles
+	janus_websockets_ht websockets; // controlling transports, websocket_conn -> websocket_conn
+	janus_handles_set handles; // handle ID -> 0x1. handle ID owned by janus_handles
 };
+
+TYPED_GHASHTABLE(janus_sessions_ht, uint64_t, struct janus_session, int64_hash, int64_eq, NULL, NULL)
+
+
 struct janus_handle { // corresponds to a conference participant
 	uint64_t id;
 	struct janus_session *session; // holds a reference
 	uint64_t room;
 };
+
+TYPED_GHASHTABLE(janus_handles_ht, uint64_t, struct janus_handle, int64_hash, int64_eq, NULL, NULL)
+
+
+TYPED_GHASHTABLE(janus_feeds_ht, uint64_t, uint64_t, int64_hash, int64_eq, g_free, g_free)
+
 struct janus_room {
 	uint64_t id;
 	str call_id;
 	int num_publishers;
 	uint64_t handle_id; // controlling handle which created the room
-	GHashTable *publishers; // handle ID -> feed ID
-	GHashTable *subscribers; // handle ID -> subscribed feed ID
-	GHashTable *feeds; // feed ID -> handle ID
+	janus_feeds_ht publishers; // handle ID -> feed ID
+	janus_feeds_ht subscribers; // handle ID -> subscribed feed ID
+	janus_feeds_ht feeds; // feed ID -> handle ID
 };
 
+TYPED_GHASHTABLE(janus_rooms_ht, uint64_t, struct janus_room, int64_hash, int64_eq, NULL, NULL)
+
+
+TYPED_GHASHTABLE(janus_tokens_ht, char, time_t, c_str_hash, c_str_equal, g_free, g_free)
+
 
 static mutex_t janus_lock;
-static GHashTable *janus_tokens; // auth tokens, currently mostly unused
-static GHashTable *janus_sessions; // session ID -> session. holds a session reference
-static GHashTable *janus_handles; // handle ID -> handle
-static GHashTable *janus_rooms; // room ID -> room
+static janus_tokens_ht janus_tokens; // auth tokens, currently mostly unused
+static janus_sessions_ht janus_sessions; // session ID -> session. holds a session reference
+static janus_handles_ht janus_handles; // handle ID -> handle
+static janus_rooms_ht janus_rooms; // room ID -> room
 
 
 static void __janus_session_free(void *p) {
 	struct janus_session *s = p;
-	if (g_hash_table_size(s->websockets) != 0)
-		ilog(LOG_WARN, "Janus session is leaking %i WS references", g_hash_table_size(s->websockets));
-	g_hash_table_destroy(s->websockets);
-	if (g_hash_table_size(s->handles) != 0)
-		ilog(LOG_WARN, "Janus session is leaking %i handle references", g_hash_table_size(s->handles));
-	g_hash_table_destroy(s->handles);
+	if (t_hash_table_size(s->websockets) != 0)
+		ilog(LOG_WARN, "Janus session is leaking %i WS references", t_hash_table_size(s->websockets));
+	t_hash_table_destroy(s->websockets);
+	if (t_hash_table_size(s->handles) != 0)
+		ilog(LOG_WARN, "Janus session is leaking %i handle references", t_hash_table_size(s->handles));
+	t_hash_table_destroy(s->handles);
 	mutex_destroy(&s->lock);
 }
 
@@ -59,7 +80,7 @@ static void __janus_session_free(void *p) {
 // XXX we have several hash tables that hold references to objs - unify all these
 static struct janus_session *janus_get_session(uint64_t id) {
 	mutex_lock(&janus_lock);
-	struct janus_session *ret = g_hash_table_lookup(janus_sessions, &id);
+	struct janus_session *ret = t_hash_table_lookup(janus_sessions, &id);
 	if (ret)
 		obj_hold(ret);
 	mutex_unlock(&janus_lock);
@@ -95,10 +116,10 @@ static uint64_t jr_str_int(JsonReader *r) {
 }
 
 
-static struct call_monologue *janus_get_monologue(uint64_t handle_id, struct call *call,
-		struct call_monologue *(*fn)(struct call *, const str *))
+static struct call_monologue *janus_get_monologue(uint64_t handle_id, call_t *call,
+		struct call_monologue *(*fn)(call_t *, const str *))
 {
-	AUTO_CLEANUP_GBUF(handle_buf);
+	g_autoptr(char) handle_buf = NULL;
 	handle_buf = g_strdup_printf("%" PRIu64, handle_id);
 	str handle_str = STR_INIT(handle_buf);
 
@@ -128,12 +149,11 @@ static void janus_send_json_sync_response(struct websocket_message *wm, JsonBuil
 static void janus_send_json_async(struct janus_session *session, JsonBuilder *builder) {
 	char *result = glib_json_print(builder);
 
-	GHashTableIter iter;
-	gpointer value;
-	g_hash_table_iter_init(&iter, session->websockets);
+	janus_websockets_ht_iter iter;
+	t_hash_table_iter_init(&iter, session->websockets);
 
-	while (g_hash_table_iter_next(&iter, NULL, &value)) {
-		struct websocket_conn *wc = value;
+	struct websocket_conn *wc;
+	while (t_hash_table_iter_next(&iter, NULL, &wc)) {
 		// lock order constraint: janus_session lock first, websocket_conn lock second
 		websocket_write_text(wc, result, true);
 	}
@@ -183,9 +203,9 @@ static const char *janus_videoroom_create(struct janus_session *session, struct
 		room->num_publishers = 3;
 	room->handle_id = handle->id; // controlling handle
 	// XXX optimise for 64-bit archs
-	room->publishers = g_hash_table_new_full(g_int64_hash, g_int64_equal, g_free, g_free);
-	room->subscribers = g_hash_table_new_full(g_int64_hash, g_int64_equal, g_free, g_free);
-	room->feeds = g_hash_table_new_full(g_int64_hash, g_int64_equal, g_free, g_free);
+	room->publishers = janus_feeds_ht_new();
+	room->subscribers = janus_feeds_ht_new();
+	room->feeds = janus_feeds_ht_new();
 
 	uint64_t room_id = 0;
 	if (json_reader_read_member(reader, "room")) {
@@ -197,7 +217,7 @@ static const char *janus_videoroom_create(struct janus_session *session, struct
 
 	if (room_id) {
 		*retcode = 512;
-		if (g_hash_table_lookup(janus_rooms, &room_id))
+		if (t_hash_table_lookup(janus_rooms, &room_id))
 			return "Requested room already exists";
 	}
 
@@ -205,11 +225,11 @@ static const char *janus_videoroom_create(struct janus_session *session, struct
 		if (!room_id)
 			room_id = janus_random();
 		room->id = room_id;
-		if (g_hash_table_lookup(janus_rooms, &room->id))
+		if (t_hash_table_lookup(janus_rooms, &room->id))
 			continue;
 		room->call_id.s = janus_call_id(room_id);
 		room->call_id.len = strlen(room->call_id.s);
-		struct call *call = call_get_or_create(&room->call_id, true);
+		call_t *call = call_get_or_create(&room->call_id, true);
 		if (!call) {
 			ilog(LOG_WARN, "Call with reserved Janus ID '" STR_FORMAT
 					"' already exists", STR_FMT(&room->call_id));
@@ -218,7 +238,7 @@ static const char *janus_videoroom_create(struct janus_session *session, struct
 		}
 		if (!call->created_from)
 			call->created_from = "janus";
-		g_hash_table_insert(janus_rooms, &room->id, room);
+		t_hash_table_insert(janus_rooms, &room->id, room);
 		rwlock_unlock_w(&call->master_lock);
 		obj_put(call);
 		break;
@@ -248,9 +268,9 @@ static const char *janus_videoroom_exists(struct janus_session *session,
 	bool exists = false;
 
 	if (room_id)
-		room = g_hash_table_lookup(janus_rooms, &room_id);
+		room = t_hash_table_lookup(janus_rooms, &room_id);
 	if (room) {
-		struct call *call = call_get(&room->call_id);
+		call_t *call = call_get(&room->call_id);
 		if (call) {
 			exists = true;
 			rwlock_unlock_w(&call->master_lock);
@@ -276,14 +296,14 @@ static const char *janus_videoroom_destroy(struct janus_session *session,
 	struct janus_room *room = NULL;
 
 	if (room_id)
-		g_hash_table_steal_extended(janus_rooms, &room_id, NULL, (void **) &room);
+		t_hash_table_steal_extended(janus_rooms, &room_id, NULL, &room);
 	*retcode = 426;
 	if (!room)
 		return "No such room";
 
 	ilog(LOG_INFO, "Destroying videoroom with ID %" PRIu64, room_id);
 
-	struct call *call = call_get(&room->call_id);
+	call_t *call = call_get(&room->call_id);
 	// XXX if call is destroyed separately, room persist -> room should be destroyed too
 	if (call) {
 		rwlock_unlock_w(&call->master_lock);
@@ -292,9 +312,9 @@ static const char *janus_videoroom_destroy(struct janus_session *session,
 	}
 
 	g_free(room->call_id.s);
-	g_hash_table_destroy(room->publishers);
-	g_hash_table_destroy(room->subscribers);
-	g_hash_table_destroy(room->feeds);
+	t_hash_table_destroy(room->publishers);
+	t_hash_table_destroy(room->subscribers);
+	t_hash_table_destroy(room->feeds);
 	g_slice_free1(sizeof(*room), room);
 
 	//XXX notify?
@@ -323,8 +343,8 @@ static void janus_add_publisher_details(JsonBuilder *builder, struct call_monolo
 			continue;
 
 		const char *codec = NULL;
-		for (GList *k = media->codecs.codec_prefs.head; k; k = k->next) {
-			struct rtp_payload_type *pt = k->data;
+		for (__auto_type k = media->codecs.codec_prefs.head; k; k = k->next) {
+			rtp_payload_type *pt = k->data;
 			codec = pt->encoding.s;
 			// XXX check codec support?
 			break;
@@ -336,59 +356,60 @@ static void janus_add_publisher_details(JsonBuilder *builder, struct call_monolo
 		json_builder_add_string_value(builder, media->type.s);
 		json_builder_set_member_name(builder, "mindex");
 		json_builder_add_int_value(builder, media->index - 1);
+
 		json_builder_set_member_name(builder, "mid");
 		if (media->media_id.s)
 			json_builder_add_string_value(builder, media->media_id.s);
 		else
 			json_builder_add_null_value(builder);
-		json_builder_set_member_name(builder, "codec");
-		if (codec)
+
+		if (!MEDIA_ISSET2(media, SEND, RECV)) {
+			json_builder_set_member_name(builder, "disabled");
+			json_builder_add_boolean_value(builder, true);
+		}
+		else if (codec) {
+			json_builder_set_member_name(builder, "codec");
 			json_builder_add_string_value(builder, codec);
-		else
-			json_builder_add_null_value(builder);
+
+			if (media->type_id == MT_AUDIO && !a_codec)
+				a_codec = codec;
+			else if (media->type_id == MT_VIDEO && !v_codec)
+				v_codec = codec;
+		}
 
 		json_builder_end_object(builder);
-
-		if (media->type_id == MT_AUDIO)
-			a_codec = codec;
-		else if (media->type_id == MT_VIDEO)
-			v_codec = codec;
 	}
 
 	json_builder_end_array(builder);
 
-	json_builder_set_member_name(builder, "audio_codec");
-	if (a_codec)
+	if (a_codec) {
+		json_builder_set_member_name(builder, "audio_codec");
 		json_builder_add_string_value(builder, a_codec);
-	else
-		json_builder_add_null_value(builder);
+	}
 
-	json_builder_set_member_name(builder, "video_codec");
-	if (v_codec)
+	if (v_codec) {
+		json_builder_set_member_name(builder, "video_codec");
 		json_builder_add_string_value(builder, v_codec);
-	else
-		json_builder_add_null_value(builder);
+	}
 
 	// TODO add "display"
 }
 
 
-static void janus_publishers_list(JsonBuilder *builder, struct call *call, struct janus_room *room,
+static void janus_publishers_list(JsonBuilder *builder, call_t *call, struct janus_room *room,
 		uint64_t feed_id)
 {
 	json_builder_begin_array(builder); // [
 
-	GHashTableIter iter;
-	gpointer key, value;
-	g_hash_table_iter_init(&iter, room->publishers);
+	janus_feeds_ht_iter iter;
+	t_hash_table_iter_init(&iter, room->publishers);
 
-	while (g_hash_table_iter_next(&iter, &key, &value)) {
-		uint64_t *feed_id_ptr = value;
+	uint64_t *feed_id_ptr, *handle_id_ptr;
+	while (t_hash_table_iter_next(&iter, &handle_id_ptr, &feed_id_ptr)) {
 		if (*feed_id_ptr == feed_id) // skip self
 			continue;
 
 		// get monologue
-		uint64_t *handle_id_ptr = key;
 		struct call_monologue *ml = janus_get_monologue(*handle_id_ptr, call, call_get_monologue);
 		if (!ml)
 			continue;
@@ -408,42 +429,48 @@ static void janus_publishers_list(JsonBuilder *builder, struct call *call, struc
 
 // global janus_lock is held
 static const char *janus_videoroom_join_sub(struct janus_handle *handle, struct janus_room *room, int *retcode,
-		uint64_t feed_id, struct call *call, GQueue *srcs)
+		uint64_t feed_id, call_t *call, subscription_q *medias)
 {
 	// does the feed actually exist? get the feed handle
 	*retcode = 512;
-	uint64_t *feed_handle = g_hash_table_lookup(room->feeds, &feed_id);
+	uint64_t *feed_handle = t_hash_table_lookup(room->feeds, &feed_id);
 	if (!feed_handle)
 		return "No such feed exists";
-	if (!g_hash_table_lookup(room->publishers, feed_handle))
+	if (!t_hash_table_lookup(room->publishers, feed_handle))
 		return "No such feed handle exists";
 
 	// handle ID points to the subscribed feed
-	g_hash_table_insert(room->subscribers, uint64_dup(handle->id), uint64_dup(feed_id));
+	t_hash_table_insert(room->subscribers, uint64_dup(handle->id), uint64_dup(feed_id));
 
 	// add the subscription
 	struct call_monologue *source_ml = janus_get_monologue(*feed_handle, call, call_get_monologue);
 	if (!source_ml)
 		return "Feed not found";
 
-	struct call_subscription *cs = g_slice_alloc0(sizeof(*cs));
-	cs->monologue = source_ml;
-	g_queue_push_tail(srcs, cs);
-
+	for (int i = 0; i < source_ml->medias->len; i++)
+	{
+		struct call_media * media = source_ml->medias->pdata[i];
+		if (!media)
+			continue;
+		add_media_to_sub_list(medias, media, source_ml);
+	}
 	return NULL;
 }
 
 
-static void janus_clear_ret_streams(GQueue *q) {
+TYPED_GQUEUE(janus_ret_streams, uint64_t);
+
+static void janus_clear_ret_streams(janus_ret_streams_q *q) {
 	uint64_t *id;
-	while ((id = g_queue_pop_head(q)))
+	while ((id = t_queue_pop_head(q)))
 		g_slice_free1(sizeof(*id), id);
 }
 
+G_DEFINE_AUTO_CLEANUP_CLEAR_FUNC(janus_ret_streams_q, janus_clear_ret_streams);
 
-static int g_int64_cmp(gconstpointer a, gconstpointer b) {
-	const uint64_t *A = a, *B = b;
-	return !(*A == *B);
+
+static int int64_cmp(const uint64_t *a, const void *b) {
+	return !(*a == *(uint64_t *) b);
 }
 
 
@@ -485,33 +512,33 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 
 	struct janus_room *room = NULL;
 	if (room_id)
-		room = g_hash_table_lookup(janus_rooms, &room_id);
+		room = t_hash_table_lookup(janus_rooms, &room_id);
 	*retcode = 426;
 	if (!room)
 		return "No such room";
 
-	AUTO_CLEANUP_NULL(struct call *call, call_unlock_release);
+	g_autoptr(call_t) call = NULL;
 	*retcode = 426;
 	call = call_get(&room->call_id);
 	if (!call)
 		return "No such room";
 
 	*retcode = 436;
-	if (!is_pub && g_hash_table_lookup(room->subscribers, &handle->id))
+	if (!is_pub && t_hash_table_lookup(room->subscribers, &handle->id))
 		return "User already exists in the room as a subscriber";
-	if (is_pub && g_hash_table_lookup(room->publishers, &handle->id))
+	if (is_pub && t_hash_table_lookup(room->publishers, &handle->id))
 		return "User already exists in the room as a publisher";
 
 	uint64_t feed_id = 0; // set for single feed IDs, otherwise remains 0
-	AUTO_CLEANUP_INIT(GString *feed_ids, __g_string_free, g_string_new("feeds ")); // for log output
-	AUTO_CLEANUP(GQueue ret_streams, janus_clear_ret_streams) = G_QUEUE_INIT; // return list for multiple subs
+	g_autoptr(GString) feed_ids = g_string_new("feeds "); // for log output
+	g_auto(janus_ret_streams_q) ret_streams = TYPED_GQUEUE_INIT; // return list for multiple subs
 
 	if (is_pub) {
 		if (json_reader_read_member(reader, "id")) {
 			feed_id = jr_str_int(reader);
 			if (!feed_id)
 				return "Invalid feed ID requested";
-			if (g_hash_table_lookup(room->feeds, &feed_id))
+			if (t_hash_table_lookup(room->feeds, &feed_id))
 				return "Feed already exists";
 		}
 		json_reader_end_member(reader);
@@ -519,19 +546,19 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 		// random feed ID?
 		while (!feed_id) {
 			feed_id = janus_random();
-			if (feed_id && g_hash_table_lookup(room->feeds, &feed_id))
+			if (feed_id && t_hash_table_lookup(room->feeds, &feed_id))
 				feed_id = 0;
 		}
 
 		// feed ID points to the handle
-		g_hash_table_insert(room->feeds, uint64_dup(feed_id), uint64_dup(handle->id));
+		t_hash_table_insert(room->feeds, uint64_dup(feed_id), uint64_dup(handle->id));
 		// handle ID points to the feed
-		g_hash_table_insert(room->publishers, uint64_dup(handle->id), uint64_dup(feed_id));
+		t_hash_table_insert(room->publishers, uint64_dup(handle->id), uint64_dup(feed_id));
 	}
 	else {
 		// subscriber
 
-		AUTO_CLEANUP(GQueue srcs, call_subscriptions_clear) = G_QUEUE_INIT;
+		g_auto(subscription_q) srms = TYPED_GQUEUE_INIT;
 
 		// get single feed ID if there is one
 		if (json_reader_read_member(reader, "feed")) {
@@ -540,7 +567,7 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 			if (!feed_id)
 				return "JSON object contains invalid 'message.feed' key";
 			const char *ret = janus_videoroom_join_sub(handle, room, retcode, feed_id,
-					call, &srcs);
+					call, &srms);
 			if (ret)
 				return ret;
 		}
@@ -572,9 +599,9 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 				// and so simply honour each unique feed ID given.
 				// TODO: fix this up
 
-				if (!g_queue_find_custom(&ret_streams, &fid, g_int64_cmp)) {
+				if (!t_queue_find_custom(&ret_streams, &fid, int64_cmp)) {
 					const char *ret = janus_videoroom_join_sub(handle, room, retcode, fid,
-						call, &srcs);
+						call, &srms);
 					if (ret)
 						return ret;
 
@@ -582,7 +609,7 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 
 					uint64_t *fidp = g_slice_alloc(sizeof(*fidp));
 					*fidp = fid;
-					g_queue_push_tail(&ret_streams, fidp);
+					t_queue_push_tail(&ret_streams, fidp);
 				}
 
 				json_reader_end_member(reader);
@@ -592,49 +619,33 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 		json_reader_end_member(reader);
 
 		*retcode = 456;
-		if (!srcs.length)
+		if (!srms.length)
 			return "No feeds to subscribe to given";
 
 		struct call_monologue *dest_ml = janus_get_monologue(handle->id, call,
 				call_get_or_create_monologue);
 
-		AUTO_CLEANUP(struct sdp_ng_flags flags, call_ng_free_flags);
+		g_auto(sdp_ng_flags) flags;
 		call_ng_flags_init(&flags, OP_REQUEST);
 
 		flags.generate_mid = 1;
 		flags.rtcp_mirror = 1;
+		flags.replace_origin_full = 1;
 
-		if (!plain_offer) {
-			// set all WebRTC-specific attributes
-			flags.transport_protocol = &transport_protocols[PROTO_UDP_TLS_RTP_SAVPF];
-			flags.ice_option = ICE_FORCE;
-			flags.trickle_ice = 1;
-			flags.rtcp_mux_offer = 1;
-			flags.rtcp_mux_require = 1;
-			flags.no_rtcp_attr = 1;
-			flags.sdes_off = 1;
-		}
+		if (!plain_offer)
+			ng_flags_webrtc(&flags);
 		else {
 			flags.transport_protocol = &transport_protocols[PROTO_RTP_AVP];
 			flags.ice_option = ICE_REMOVE;
 			flags.rtcp_mux_demux = 1;
 		}
 
-		int ret = monologue_subscribe_request(&srcs, dest_ml, &flags);
+		int ret = monologue_subscribe_request(&srms, dest_ml, &flags);
 		if (ret)
 			return "Subscribe error";
 
-		// create SDP: if there's only one subscription, we can use the original
-		// SDP, otherwise we generate a new one
-		if (srcs.length == 1) {
-			struct call_subscription *cs = srcs.head->data;
-			struct call_monologue *source_ml = cs->monologue;
-			struct sdp_chopper *chopper = sdp_chopper_new(&source_ml->last_in_sdp);
-			ret = sdp_replace(chopper, &source_ml->last_in_sdp_parsed, dest_ml, &flags);
-			sdp_chopper_destroy_ret(chopper, jsep_sdp_out);
-		}
-		else
-			ret = sdp_create(jsep_sdp_out, dest_ml, &flags);
+		/* create SDP */
+		ret = sdp_create(jsep_sdp_out, dest_ml, &flags);
 
 		if (!dest_ml->janus_session)
 			dest_ml->janus_session = obj_get(session);
@@ -643,6 +654,7 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 
 		if (ret)
 			return "Error generating SDP";
+
 		*jsep_type_out = "offer";
 	}
 
@@ -686,7 +698,7 @@ static const char *janus_videoroom_join(struct websocket_message *wm, struct jan
 			json_builder_set_member_name(builder, "streams");
 			json_builder_begin_array(builder);
 			uint64_t idx = 0;
-			for (GList *l = ret_streams.head; l; l = l->next) {
+			for (__auto_type l = ret_streams.head; l; l = l->next) {
 				uint64_t *fidp = l->data;
 				json_builder_begin_object(builder);
 				json_builder_set_member_name(builder, "mindex");
@@ -735,21 +747,20 @@ static void janus_notify_publishers(uint64_t room_id, uint64_t except, void *ptr
 		void (*callback)(JsonBuilder *event, void *ptr, uint64_t u64, struct janus_room *room,
 			uint64_t publisher_feed))
 {
-	struct janus_room *room = g_hash_table_lookup(janus_rooms, &room_id);
+	struct janus_room *room = t_hash_table_lookup(janus_rooms, &room_id);
 	if (!room)
 		return;
 
-	GHashTableIter iter;
-	gpointer key, value;
-	g_hash_table_iter_init(&iter, room->publishers);
+	janus_feeds_ht_iter iter;
+	t_hash_table_iter_init(&iter, room->publishers);
 
-	while (g_hash_table_iter_next(&iter, &key, &value)) {
-		uint64_t *handle_id = key;
+	uint64_t *handle_id, *feed_id;
+	while (t_hash_table_iter_next(&iter, &handle_id, &feed_id)) {
 		if (*handle_id == except)
 			continue;
 
 		// look up the handle and determine which session it belongs to
-		struct janus_handle *handle = g_hash_table_lookup(janus_handles, handle_id);
+		struct janus_handle *handle = t_hash_table_lookup(janus_handles, handle_id);
 		if (!handle)
 			continue;
 		if (!handle->session)
@@ -757,8 +768,6 @@ static void janus_notify_publishers(uint64_t room_id, uint64_t except, void *ptr
 
 		// send to the handle's session
 
-		uint64_t *feed_id = value;
-
 		JsonBuilder *event = json_builder_new();
 		json_builder_begin_object(event); // {
 		json_builder_set_member_name(event, "janus");
@@ -824,18 +833,16 @@ static const char *janus_videoroom_configure(struct websocket_message *wm, struc
 	if (handle->room != room_id)
 		return "Not in the room";
 
-	AUTO_CLEANUP_NULL(struct call *call, call_unlock_release);
-
-	struct janus_room *room = g_hash_table_lookup(janus_rooms, &room_id);
+	struct janus_room *room = t_hash_table_lookup(janus_rooms, &room_id);
 	*retcode = 426;
 	if (!room)
 		return "No such room";
-	call = call_get(&room->call_id);
+	g_autoptr(call_t) call = call_get(&room->call_id);
 	// XXX if call is destroyed separately, room persists -> room should be destroyed too
 	if (!call)
 		return "No such room";
 	*retcode = 512;
-	if (!g_hash_table_lookup(room->publishers, &handle->id))
+	if (!t_hash_table_lookup(room->publishers, &handle->id))
 		return "Not a publisher";
 
 	struct call_monologue *ml = NULL;
@@ -844,11 +851,11 @@ static const char *janus_videoroom_configure(struct websocket_message *wm, struc
 		if (strcmp(jsep_type, "offer"))
 			return "Not an offer";
 
-		AUTO_CLEANUP(str sdp_in, str_free_dup) = STR_INIT_DUP(jsep_sdp);
+		g_auto(str) sdp_in = STR_INIT_DUP(jsep_sdp);
 
-		AUTO_CLEANUP(struct sdp_ng_flags flags, call_ng_free_flags);
-		AUTO_CLEANUP(GQueue parsed, sdp_free) = G_QUEUE_INIT;
-		AUTO_CLEANUP(GQueue streams, sdp_streams_free) = G_QUEUE_INIT;
+		g_auto(sdp_ng_flags) flags;
+		g_auto(sdp_sessions_q) parsed = TYPED_GQUEUE_INIT;
+		g_auto(sdp_streams_q) streams = TYPED_GQUEUE_INIT;
 		call_ng_flags_init(&flags, OP_PUBLISH);
 		*retcode = 512;
 		if (sdp_parse(&sdp_in, &parsed, &flags))
@@ -860,6 +867,7 @@ static const char *janus_videoroom_configure(struct websocket_message *wm, struc
 
 		// accept unsupported codecs if necessary
 		flags.accept_any = 1;
+		flags.replace_origin_full = 1;
 
 		int ret = monologue_publish(ml, &streams, &flags);
 		if (ret)
@@ -867,7 +875,7 @@ static const char *janus_videoroom_configure(struct websocket_message *wm, struc
 
 		// XXX check there's only one audio and one video stream?
 
-		AUTO_CLEANUP(str sdp_out, str_free_dup) = STR_NULL;
+		g_auto(str) sdp_out = STR_NULL;
 		ret = sdp_create(&sdp_out, ml, &flags);
 		if (ret)
 			return "Publish error";
@@ -948,11 +956,11 @@ static const char *janus_videoroom_start(struct websocket_message *wm, struct ja
 	if (strcmp(jsep_type, "answer"))
 		return "Not an answer";
 
-	AUTO_CLEANUP(str sdp_in, str_free_dup) = STR_INIT_DUP(jsep_sdp);
+	g_auto(str) sdp_in = STR_INIT_DUP(jsep_sdp);
 
-	AUTO_CLEANUP(struct sdp_ng_flags flags, call_ng_free_flags);
-	AUTO_CLEANUP(GQueue parsed, sdp_free) = G_QUEUE_INIT;
-	AUTO_CLEANUP(GQueue streams, sdp_streams_free) = G_QUEUE_INIT;
+	g_auto(sdp_ng_flags) flags;
+	g_auto(sdp_sessions_q) parsed = TYPED_GQUEUE_INIT;
+	g_auto(sdp_streams_q) streams = TYPED_GQUEUE_INIT;
 	call_ng_flags_init(&flags, OP_PUBLISH);
 	*retcode = 512;
 	if (sdp_parse(&sdp_in, &parsed, &flags))
@@ -960,22 +968,20 @@ static const char *janus_videoroom_start(struct websocket_message *wm, struct ja
 	if (sdp_streams(&parsed, &streams, &flags))
 		return "Incomplete SDP specification";
 
-	AUTO_CLEANUP_NULL(struct call *call, call_unlock_release);
-
-	struct janus_room *room = g_hash_table_lookup(janus_rooms, &room_id);
+	struct janus_room *room = t_hash_table_lookup(janus_rooms, &room_id);
 	*retcode = 426;
 	if (!room)
 		return "No such room";
-	call = call_get(&room->call_id);
+	g_autoptr(call_t) call = call_get(&room->call_id);
 	if (!call)
 		return "No such room";
 	*retcode = 456;
-	uint64_t *feed_id = g_hash_table_lookup(room->subscribers, &handle->id);
+	uint64_t *feed_id = t_hash_table_lookup(room->subscribers, &handle->id);
 	if (!feed_id)
 		return "Not a subscriber";
 
 	*retcode = 512;
-	uint64_t *feed_handle = g_hash_table_lookup(room->feeds, feed_id);
+	uint64_t *feed_handle = t_hash_table_lookup(room->feeds, feed_id);
 	if (!feed_handle)
 		return "No such feed exists";
 
@@ -1021,17 +1027,16 @@ static const char *janus_videoroom_unpublish(struct websocket_message *wm, struc
 
 	struct janus_room *room = NULL;
 	if (room_id)
-		room = g_hash_table_lookup(janus_rooms, &room_id);
+		room = t_hash_table_lookup(janus_rooms, &room_id);
 	*retcode = 426;
 	if (!room)
 		return "No such room";
 
-	AUTO_CLEANUP_NULL(struct call *call, call_unlock_release);
-	call = call_get(&room->call_id);
+	g_autoptr(call_t) call = call_get(&room->call_id);
 	if (!call)
 		return "No such room";
 
-	uint64_t *feed_id = g_hash_table_lookup(room->publishers, &handle->id);
+	uint64_t *feed_id = t_hash_table_lookup(room->publishers, &handle->id);
 	*retcode = 512;
 	if (!feed_id)
 		return "Not a publisher";
@@ -1078,7 +1083,7 @@ static const char *janus_videoroom(struct websocket_message *wm, struct janus_se
 	const char *req = json_reader_get_string_value(reader);
 	if (!req)
 		goto err;
-	str req_str = STR_INIT((char*) req);
+	str req_str = STR_INIT(req);
 	json_reader_end_member(reader);
 
 	switch (__csh_lookup(&req_str)) {
@@ -1147,7 +1152,7 @@ static const char *janus_add_token(JsonReader *reader, JsonBuilder *builder, boo
 	time_t *now = g_malloc(sizeof(*now));
 	*now = rtpe_now.tv_sec;
 	mutex_lock(&janus_lock);
-	g_hash_table_replace(janus_tokens, g_strdup(token), now);
+	t_hash_table_replace(janus_tokens, g_strdup(token), now);
 	mutex_unlock(&janus_lock);
 
 	json_builder_set_member_name(builder, "data");
@@ -1175,21 +1180,21 @@ static const char *janus_create(JsonReader *reader, JsonBuilder *builder, struct
 	mutex_init(&session->lock);
 	mutex_lock(&session->lock); // not really necessary but Coverity complains
 	session->last_act = rtpe_now.tv_sec;
-	session->websockets = g_hash_table_new(g_direct_hash, g_direct_equal);
-	session->handles = g_hash_table_new(g_int64_hash, g_int64_equal);
+	session->websockets = janus_websockets_ht_new();
+	session->handles = janus_handles_set_new();
 
-	g_hash_table_insert(session->websockets, wm->wc, wm->wc);
+	t_hash_table_insert(session->websockets, wm->wc, wm->wc);
 
 	do {
 		while (!session_id)
 			session_id = janus_random();
 
 		mutex_lock(&janus_lock);
-		if (g_hash_table_lookup(janus_sessions, &session_id))
+		if (t_hash_table_lookup(janus_sessions, &session_id))
 			session_id = 0; // pick a random one
 		else {
 			session->id = session_id;
-			g_hash_table_insert(janus_sessions, &session->id, obj_get(session));
+			t_hash_table_insert(janus_sessions, &session->id, obj_get(session));
 		}
 		mutex_unlock(&janus_lock);
 	}
@@ -1214,7 +1219,7 @@ static const char *janus_create(JsonReader *reader, JsonBuilder *builder, struct
 
 void janus_detach_websocket(struct janus_session *session, struct websocket_conn *wc) {
 	LOCK(&session->lock);
-	g_hash_table_remove(session->websockets, wc);
+	t_hash_table_remove(session->websockets, wc);
 }
 
 
@@ -1313,16 +1318,16 @@ static const char *janus_attach(JsonReader *reader, JsonBuilder *builder, struct
 	uint64_t handle_id = 0;
 	while (1) {
 		handle_id = handle->id = janus_random();
-		if (g_hash_table_lookup(janus_handles, &handle->id))
+		if (t_hash_table_lookup(janus_handles, &handle->id))
 			continue;
-		g_hash_table_insert(janus_handles, &handle->id, handle);
+		t_hash_table_insert(janus_handles, &handle->id, handle);
 		break;
 	}
 	mutex_unlock(&janus_lock);
 
 	mutex_lock(&session->lock);
-	assert(g_hash_table_lookup(session->handles, &handle_id) == NULL);
-	g_hash_table_insert(session->handles, &handle->id, (void *) 0x1);
+	assert(t_hash_table_lookup(session->handles, &handle_id) == NULL);
+	t_hash_table_insert(session->handles, &handle->id, (void *) 0x1);
 	mutex_unlock(&session->lock);
 
 	json_builder_set_member_name(builder, "data");
@@ -1347,17 +1352,17 @@ static void janus_destroy_handle(struct janus_handle *handle) {
 	if (!room_id)
 		return;
 
-	struct janus_room *room = g_hash_table_lookup(janus_rooms, &room_id);
+	struct janus_room *room = t_hash_table_lookup(janus_rooms, &room_id);
 	if (!room)
 		return;
 
-	uint64_t *feed = g_hash_table_lookup(room->publishers, &handle_id);
+	uint64_t *feed = t_hash_table_lookup(room->publishers, &handle_id);
 	if (feed) {
 		// was a publisher - send notifies
 		janus_notify_publishers(room_id, handle_id, NULL, *feed, janus_notify_publishers_unpublished);
 		janus_notify_publishers(room_id, handle_id, NULL, *feed, janus_notify_publishers_leaving);
 
-		struct call *call = call_get(&room->call_id);
+		call_t *call = call_get(&room->call_id);
 		if (call) {
 			// remove publisher monologue
 			struct call_monologue *ml = janus_get_monologue(handle_id, call, call_get_monologue);
@@ -1368,13 +1373,13 @@ static void janus_destroy_handle(struct janus_handle *handle) {
 			obj_put(call);
 		}
 
-		g_hash_table_remove(room->publishers, &handle_id);
+		t_hash_table_remove(room->publishers, &handle_id);
 		feed = NULL;
 	}
 
-	if (g_hash_table_remove(room->subscribers, &handle_id)) {
+	if (t_hash_table_remove(room->subscribers, &handle_id)) {
 		// was a subscriber
-		struct call *call = call_get(&room->call_id);
+		call_t *call = call_get(&room->call_id);
 		if (call) {
 			// remove subscriber monologue
 			struct call_monologue *ml = janus_get_monologue(handle_id, call, call_get_monologue);
@@ -1404,7 +1409,7 @@ static const char *janus_detach(struct websocket_message *wm, JsonReader *reader
 	{
 		LOCK(&session->lock);
 
-		bool exists = g_hash_table_remove(session->handles, &handle_id);
+		bool exists = t_hash_table_remove(session->handles, &handle_id);
 
 		*retcode = 463;
 		if (!exists)
@@ -1414,13 +1419,13 @@ static const char *janus_detach(struct websocket_message *wm, JsonReader *reader
 	LOCK(&janus_lock);
 
 	struct janus_handle *handle = NULL;
-	g_hash_table_steal_extended(janus_handles, &handle_id, NULL, (void **) &handle);
+	t_hash_table_steal_extended(janus_handles, &handle_id, NULL, &handle);
 
 	*retcode = 463;
 	if (!handle)
 		return "Could not detach handle from plugin";
 	if (handle->session != session) {
-		g_hash_table_insert(janus_handles, &handle->id, handle);
+		t_hash_table_insert(janus_handles, &handle->id, handle);
 		return "Invalid session/handle association";
 	}
 
@@ -1432,13 +1437,12 @@ static const char *janus_detach(struct websocket_message *wm, JsonReader *reader
 
 // janus_lock must be held
 static void janus_session_cleanup(struct janus_session *session) {
-	GHashTableIter iter;
-	g_hash_table_iter_init(&iter, session->handles);
-	gpointer key;
-	while (g_hash_table_iter_next(&iter, &key, NULL)) {
-		uint64_t *handle_id = key;
+	janus_handles_set_iter iter;
+	t_hash_table_iter_init(&iter, session->handles);
+	uint64_t *handle_id;
+	while (t_hash_table_iter_next(&iter, &handle_id, NULL)) {
 		struct janus_handle *handle = NULL;
-		g_hash_table_steal_extended(janus_handles, handle_id, NULL, (void **) &handle);
+		t_hash_table_steal_extended(janus_handles, handle_id, NULL, &handle);
 		if (!handle) // bug?
 			continue;
 		janus_destroy_handle(handle);
@@ -1457,7 +1461,7 @@ static const char *janus_destroy(struct websocket_message *wm, JsonReader *reade
 	LOCK(&janus_lock);
 
 	struct janus_session *ht_session = NULL;
-	g_hash_table_steal_extended(janus_sessions, &session->id, NULL, (void **) &ht_session);
+	t_hash_table_steal_extended(janus_sessions, &session->id, NULL, &ht_session);
 	if (ht_session != session)
 		return "Sesssion ID not found"; // already removed/destroyed
 
@@ -1507,11 +1511,11 @@ static const char *janus_message(struct websocket_message *wm, JsonReader *reade
 	json_builder_begin_object(builder); // {
 
 	char *jsep_type_out = NULL;
-	str jsep_sdp_out = STR_NULL;
+	g_auto(str) jsep_sdp_out = STR_NULL;
 
 	LOCK(&janus_lock);
 
-	struct janus_handle *handle = g_hash_table_lookup(janus_handles, &handle_id);
+	struct janus_handle *handle = t_hash_table_lookup(janus_handles, &handle_id);
 
 	const char *err = NULL;
 	if (!handle || handle->session != session) {
@@ -1536,8 +1540,6 @@ static const char *janus_message(struct websocket_message *wm, JsonReader *reade
 		json_builder_end_object(builder); // }
 	}
 
-	str_free_dup(&jsep_sdp_out);
-
 	return err;
 
 }
@@ -1597,19 +1599,19 @@ static const char *janus_trickle(JsonReader *reader, struct janus_session *sessi
 
 	// fetch call
 
-	AUTO_CLEANUP_GBUF(call_id);
-	AUTO_CLEANUP_NULL(struct call *call, call_unlock_release);
+	g_autoptr(char) call_id = NULL;
+	g_autoptr(call_t) call = NULL;
 	{
 		LOCK(&janus_lock);
 
-		struct janus_handle *handle = g_hash_table_lookup(janus_handles, &handle_id);
+		struct janus_handle *handle = t_hash_table_lookup(janus_handles, &handle_id);
 
 		if (!handle || !handle->room || handle->session != session)
 			return "Unhandled request method";
 
 		call_id = janus_call_id(handle->room);
 
-		struct janus_room *room = g_hash_table_lookup(janus_rooms, &handle->room);
+		struct janus_room *room = t_hash_table_lookup(janus_rooms, &handle->room);
 
 		if (room)
 			call = call_get(&room->call_id);
@@ -1622,16 +1624,16 @@ static const char *janus_trickle(JsonReader *reader, struct janus_session *sessi
 	*successp = "ack";
 
 	// top-level structures first, with auto cleanup
-	AUTO_CLEANUP(GQueue streams, sdp_streams_free) = G_QUEUE_INIT;
-	AUTO_CLEANUP(struct ng_buffer *ngbuf, ng_buffer_auto_release) = ng_buffer_new(NULL);
-	AUTO_CLEANUP(struct sdp_ng_flags flags, call_ng_free_flags);
+	g_auto(sdp_streams_q) streams = TYPED_GQUEUE_INIT;
+	g_autoptr(ng_buffer) ngbuf = ng_buffer_new(NULL);
+	g_auto(sdp_ng_flags) flags;
 	call_ng_flags_init(&flags, OP_OTHER);
 
 	// then the contained structures, and add them in
 	struct stream_params *sp = g_slice_alloc0(sizeof(*sp));
-	g_queue_push_tail(&streams, sp);
+	t_queue_push_tail(&streams, sp);
 	struct ice_candidate *cand = g_slice_alloc0(sizeof(*cand));
-	g_queue_push_tail(&sp->ice_candidates, cand);
+	t_queue_push_tail(&sp->ice_candidates, cand);
 
 	// allocate and parse candidate
 	str cand_str;
@@ -1650,7 +1652,7 @@ static const char *janus_trickle(JsonReader *reader, struct janus_session *sessi
 	// set required signalling flags
 	flags.fragment = 1;
 
-	AUTO_CLEANUP_GBUF(handle_buf);
+	g_autoptr(char) handle_buf = NULL;
 	handle_buf = g_strdup_printf("%" PRIu64, handle_id);
 	bencode_strdup_str(&ngbuf->buffer, &flags.from_tag, handle_buf);
 	bencode_strdup_str(&ngbuf->buffer, &flags.call_id, call_id);
@@ -1785,7 +1787,7 @@ static const char *websocket_janus_process_json(struct websocket_message *wm,
 
 	ilog(LOG_DEBUG, "Processing '%s' type Janus message", janus_cmd);
 
-	str janus_cmd_str = STR_INIT((char*) janus_cmd);
+	str janus_cmd_str = STR_INIT(janus_cmd);
 
 	err = NULL;
 
@@ -1944,14 +1946,14 @@ const char *websocket_janus_post(struct websocket_message *wm) {
 
 	// parse out session ID and handle ID if given
 	str s;
-	if (str_token_sep(&s, &uri, '/'))
+	if (!str_token_sep(&s, &uri, '/'))
 		goto done;
 	if (str_cmp(&s, "janus"))
 		goto done;
-	if (str_token_sep(&s, &uri, '/'))
+	if (!str_token_sep(&s, &uri, '/'))
 		goto done;
 	session_id = str_to_ui(&s, 0);
-	if (str_token_sep(&s, &uri, '/'))
+	if (!str_token_sep(&s, &uri, '/'))
 		goto done;
 	handle_id = str_to_ui(&s, 0);
 
@@ -1962,16 +1964,16 @@ done:
 
 void janus_init(void) {
 	mutex_init(&janus_lock);
-	janus_tokens = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
-	janus_sessions = g_hash_table_new(g_int64_hash, g_int64_equal);
-	janus_handles = g_hash_table_new(g_int64_hash, g_int64_equal);
-	janus_rooms = g_hash_table_new(g_int64_hash, g_int64_equal);
+	janus_tokens = janus_tokens_ht_new();
+	janus_sessions = janus_sessions_ht_new();
+	janus_handles = janus_handles_ht_new();
+	janus_rooms = janus_rooms_ht_new();
 	// XXX timer thread to clean up orphaned sessions
 }
 void janus_free(void) {
 	mutex_destroy(&janus_lock);
-	g_hash_table_destroy(janus_tokens);
-	g_hash_table_destroy(janus_sessions);
-	g_hash_table_destroy(janus_handles);
-	g_hash_table_destroy(janus_rooms);
+	t_hash_table_destroy(janus_tokens);
+	t_hash_table_destroy(janus_sessions);
+	t_hash_table_destroy(janus_handles);
+	t_hash_table_destroy(janus_rooms);
 }
