@@ -19,6 +19,9 @@
 #ifdef HAVE_MQTT
 #include <mosquitto.h>
 #endif
+#ifndef WITHOUT_NFTABLES
+#include <linux/netfilter.h>
+#endif
 
 #include "poller.h"
 #include "control_tcp.h"
@@ -53,11 +56,20 @@
 #include "codec.h"
 #include "mqtt.h"
 #include "janus.h"
+#include "nftables.h"
+#include "bufferpool.h"
+#include "log_funcs.h"
+#include "uring.h"
 
 
 
-struct poller *rtpe_poller;
-struct poller_map *rtpe_poller_map;
+struct poller **rtpe_pollers;
+struct poller *rtpe_control_poller;
+static unsigned int num_rtpe_pollers;
+static unsigned int num_poller_threads;
+unsigned int num_media_pollers;
+unsigned int rtpe_poller_rr_iter;
+
 struct rtpengine_config initial_rtpe_config;
 
 static GQueue rtpe_tcp = G_QUEUE_INIT;
@@ -66,6 +78,7 @@ static GQueue rtpe_cli = G_QUEUE_INIT;
 
 GQueue rtpe_control_ng = G_QUEUE_INIT;
 GQueue rtpe_control_ng_tcp = G_QUEUE_INIT;
+struct bufferpool *shm_bufferpool;
 
 struct rtpengine_config rtpe_config = {
 	// non-zero defaults
@@ -74,9 +87,10 @@ struct rtpengine_config rtpe_config = {
 	.delete_delay = 30,
 	.redis_subscribed_keyspaces = G_QUEUE_INIT,
 	.redis_expires_secs = 86400,
-	.interfaces = G_QUEUE_INIT,
+	.interfaces = TYPED_GQUEUE_INIT,
 	.homer_protocol = SOCK_DGRAM,
 	.homer_id = 2001,
+	.homer_ng_capt_proto = 0x3d, // first available value in HEP proto specification
 	.port_min = 30000,
 	.port_max = 40000,
 	.redis_db = -1,
@@ -97,11 +111,14 @@ struct rtpengine_config rtpe_config = {
 	.mqtt_keepalive = 30,
 	.mqtt_publish_interval = 5000,
 	.dtmf_digit_delay = 2500,
+	.rtcp_interval = 5000,
 	.common = {
 		.log_levels = {
 			[log_level_index_internals] = -1,
 		},
 	},
+	.max_recv_iters = MAX_RECV_ITERS,
+	.kernel_player_media = 128,
 };
 
 static void sighandler(gpointer x) {
@@ -251,7 +268,7 @@ static void __resolve_ifname(char *s, GQueue *addrs) {
 	freeaddrinfo(res);
 }
 
-static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
+static int if_addr_parse(intf_config_q *q, char *s, struct ifaddrs *ifas) {
 	str name;
 	char *c;
 	sockaddr_t *addr, adv;
@@ -259,11 +276,20 @@ static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
 	struct intf_config *ifa;
 
 	/* name */
-	c = strchr(s, '/');
+	c = strpbrk(s, "/=");
 	if (c) {
+		char cc = *c;
 		*c++ = 0;
 		str_init(&name, s);
 		s = c;
+		if (cc == '=') {
+			// foo=bar
+			ifa = g_slice_alloc0(sizeof(*ifa));
+			str_init_dup_str(&ifa->name, &name);
+			str_init_dup(&ifa->alias, s);
+			t_queue_push_tail(q, ifa);
+			return 0;
+		}
 	}
 	else
 		str_init(&name, "default");
@@ -325,7 +351,7 @@ static int if_addr_parse(GQueue *q, char *s, struct ifaddrs *ifas) {
 		ifa->name_rr_spec = ifa->name;
 		str_token(&ifa->name_base, &ifa->name_rr_spec, ':'); // sets name_rr_spec to null string if no ':' found
 
-		g_queue_push_tail(q, ifa);
+		t_queue_push_tail(q, ifa);
 
 		g_slice_free1(sizeof(*addr), addr);
 	}
@@ -444,52 +470,65 @@ static void release_listeners(GQueue *q) {
 
 
 static void options(int *argc, char ***argv) {
-	AUTO_CLEANUP_GVBUF(if_a);
-	AUTO_CLEANUP_GVBUF(ks_a);
+	g_autoptr(char_p) if_a = NULL;
+	g_autoptr(char_p) ks_a = NULL;
 	unsigned long uint_keyspace_db;
 	str str_keyspace_db;
 	char **iter;
-	AUTO_CLEANUP_GVBUF(listenps);
-	AUTO_CLEANUP_GVBUF(listenudps);
-	AUTO_CLEANUP_GVBUF(listenngs);
-	AUTO_CLEANUP_GVBUF(listenngtcps);
-	AUTO_CLEANUP_GVBUF(listencli);
-	AUTO_CLEANUP_GBUF(graphitep);
-	AUTO_CLEANUP_GBUF(graphite_prefix_s);
-	AUTO_CLEANUP_GBUF(redisps);
-	AUTO_CLEANUP_GBUF(redisps_write);
-	AUTO_CLEANUP_GBUF(log_facility_cdr_s);
-	AUTO_CLEANUP_GBUF(log_facility_rtcp_s);
-	AUTO_CLEANUP_GBUF(log_facility_dtmf_s);
-	AUTO_CLEANUP_GBUF(log_format);
+	g_autoptr(char_p) listenps = NULL;
+	g_autoptr(char_p) listenudps = NULL;
+	g_autoptr(char_p) listenngs = NULL;
+	g_autoptr(char_p) listenngtcps = NULL;
+	g_autoptr(char_p) listencli = NULL;
+	g_autoptr(char) graphitep = NULL;
+	g_autoptr(char) graphite_prefix_s = NULL;
+	g_autoptr(char) redisps = NULL;
+	g_autoptr(char) redisps_write = NULL;
+	g_autoptr(char) log_facility_cdr_s = NULL;
+	g_autoptr(char) log_facility_rtcp_s = NULL;
+	g_autoptr(char) log_facility_dtmf_s = NULL;
+	g_autoptr(char) log_format = NULL;
 	bool sip_source = false;
-	AUTO_CLEANUP_GBUF(homerp);
-	AUTO_CLEANUP_GBUF(homerproto);
+	g_autoptr(char) homerp = NULL;
+	g_autoptr(char) homerproto = NULL;
 	char *endptr;
 	bool codecs = false;
 	double max_load = 0;
 	double max_cpu = 0;
-	AUTO_CLEANUP_GBUF(dtmf_udp_ep);
-	AUTO_CLEANUP_GBUF(endpoint_learning);
-	AUTO_CLEANUP_GBUF(dtls_sig);
+	g_autoptr(char) dtmf_udp_ep = NULL;
+	g_autoptr(char) endpoint_learning = NULL;
+	g_autoptr(char) dtls_sig = NULL;
 	double silence_detect = 0;
-	AUTO_CLEANUP_GVBUF(cn_payload);
-	AUTO_CLEANUP_GVBUF(dtx_cn_params);
+	g_autoptr(char_p) cn_payload = NULL;
+	g_autoptr(char_p) dtx_cn_params = NULL;
 	bool debug_srtp = false;
-	AUTO_CLEANUP_GBUF(amr_dtx);
+	g_autoptr(char) amr_dtx = NULL;
 #ifdef HAVE_MQTT
-	AUTO_CLEANUP_GBUF(mqtt_publish_scope);
+	g_autoptr(char) mqtt_publish_scope = NULL;
+#endif
+	g_autoptr(char) mos = NULL;
+	g_autoptr(char) dcc = NULL;
+	g_autoptr(char) use_audio_player = NULL;
+	g_autoptr(char) control_pmtu = NULL;
+#ifndef WITHOUT_NFTABLES
+	bool nftables_start = false;
+	bool nftables_stop = false;
+	bool nftables_status = false;
+	g_autoptr(char) nftables_family = NULL;
 #endif
-	AUTO_CLEANUP_GBUF(mos);
-	AUTO_CLEANUP_GBUF(dcc);
-	AUTO_CLEANUP_GBUF(use_audio_player);
-	AUTO_CLEANUP_GBUF(control_pmtu);
-
-	rwlock_lock_w(&rtpe_config.config_lock);
 
 	GOptionEntry e[] = {
 		{ "table",	't', 0, G_OPTION_ARG_INT,	&rtpe_config.kernel_table,		"Kernel table to use",		"INT"		},
 		{ "no-fallback",'F', 0, G_OPTION_ARG_NONE,	&rtpe_config.no_fallback,	"Only start when kernel module is available", NULL },
+#ifndef WITHOUT_NFTABLES
+		{ "nftables-chain",0,0, G_OPTION_ARG_STRING,	&rtpe_config.nftables_chain,	"Name of nftables chain to manage", "STR" },
+		{ "nftables-base-chain",0,0, G_OPTION_ARG_STRING,&rtpe_config.nftables_base_chain,"Name of nftables base chain to use", "STR" },
+		{ "nftables-append",0,0, G_OPTION_ARG_NONE,	&rtpe_config.nftables_append,	"Append instead of prepend created rules", NULL },
+		{ "nftables-family",0,0, G_OPTION_ARG_STRING,	&nftables_family,		"Address family/ies to manage via nftables", "ip|ip6|ip,ip6" },
+		{ "nftables-start",0,0, G_OPTION_ARG_NONE,	&nftables_start,		"Just add nftables rules and exit", NULL },
+		{ "nftables-stop",0, 0, G_OPTION_ARG_NONE,	&nftables_stop,			"Just remove nftables rules and exit", NULL },
+		{ "nftables-status",0, 0, G_OPTION_ARG_NONE,	&nftables_status,		"Check nftables rules, print result and exit", NULL },
+#endif
 		{ "interface",	'i', 0, G_OPTION_ARG_STRING_ARRAY,&if_a,	"Local interface for RTP",	"[NAME/]IP[!IP]"},
 		{ "save-interface-ports",'S', 0, G_OPTION_ARG_NONE,	&rtpe_config.save_interface_ports,	"Bind ports only on first available interface of desired family", NULL },
 		{ "subscribe-keyspace", 'k', 0, G_OPTION_ARG_STRING_ARRAY,&ks_a,	"Subscription keyspace list",	"INT INT ..."},
@@ -542,6 +581,9 @@ static void options(int *argc, char ***argv) {
 		{ "xmlrpc-format",'x', 0, G_OPTION_ARG_INT,	&rtpe_config.fmt,	"XMLRPC timeout request format to use. 0: SEMS DI, 1: call-id only, 2: Kamailio",	"INT"	},
 		{ "num-threads",  0, 0, G_OPTION_ARG_INT,	&rtpe_config.num_threads,	"Number of worker threads to create",	"INT"	},
 		{ "media-num-threads",  0, 0, G_OPTION_ARG_INT,	&rtpe_config.media_num_threads,	"Number of worker threads for media playback",	"INT"	},
+#ifdef WITH_TRANSCODING
+		{ "codec-num-threads",  0, 0, G_OPTION_ARG_INT,	&rtpe_config.codec_num_threads,	"Number of transcoding threads for asynchronous operation",	"INT"	},
+#endif
 		{ "delete-delay",  'd', 0, G_OPTION_ARG_INT,    &rtpe_config.delete_delay,  "Delay for deleting a session from memory.",    "INT"   },
 		{ "sip-source",  0,  0, G_OPTION_ARG_NONE,	&sip_source,	"Use SIP source address by default",	NULL	},
 		{ "dtls-passive", 0, 0, G_OPTION_ARG_NONE,	&dtls_passive_def,"Always prefer DTLS passive role",	NULL	},
@@ -552,6 +594,9 @@ static void options(int *argc, char ***argv) {
 		{ "homer",	0,  0, G_OPTION_ARG_STRING,	&homerp,	"Address of Homer server for RTCP stats","IP46|HOSTNAME:PORT"},
 		{ "homer-protocol",0,0,G_OPTION_ARG_STRING,	&homerproto,	"Transport protocol for Homer (default udp)",	"udp|tcp"	},
 		{ "homer-id",	0,  0, G_OPTION_ARG_INT,	&rtpe_config.homer_id,	"'Capture ID' to use within the HEP protocol", "INT"	},
+		{ "homer-disable-rtcp-stats", 0, 0, G_OPTION_ARG_NONE,	&rtpe_config.homer_rtcp_off,	"Disable RTCP stats tracing to Homer (enabled by default if homer server enabled)", NULL	},
+		{ "homer-enable-ng", 0, 0, G_OPTION_ARG_NONE,	&rtpe_config.homer_ng_on,	"Enable NG tracing to Homer", NULL	},
+		{ "homer-ng-capture-proto", 0, 0, G_OPTION_ARG_INT,	&rtpe_config.homer_ng_capt_proto,	"'Capture protocol type' to use within the HEP protocol (default is 0x3d). Further used by the Homer capture and UI.", "UINT8"	},
 		{ "recording-dir", 0, 0, G_OPTION_ARG_STRING,	&rtpe_config.spooldir,	"Directory for storing pcap and metadata files", "FILE"	},
 		{ "recording-method",0, 0, G_OPTION_ARG_STRING,	&rtpe_config.rec_method,	"Strategy for call recording",		"pcap|proc|all"	},
 		{ "recording-format",0, 0, G_OPTION_ARG_STRING,	&rtpe_config.rec_format,	"File format for stored pcap files",	"raw|eth"	},
@@ -598,6 +643,8 @@ static void options(int *argc, char ***argv) {
 		{ "silence-detect",0,0,	G_OPTION_ARG_DOUBLE,	&silence_detect,	"Audio level threshold in percent for silence detection","FLOAT"},
 		{ "cn-payload",0,0,	G_OPTION_ARG_STRING_ARRAY,&cn_payload,		"Comfort noise parameters to replace silence with","INT INT INT ..."},
 		{ "player-cache",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.player_cache,"Cache media files for playback in memory",NULL},
+		{ "kernel-player",0,0,	G_OPTION_ARG_INT,	&rtpe_config.kernel_player,"Max number of kernel media player streams","INT"},
+		{ "kernel-player-media",0,0,G_OPTION_ARG_INT,	&rtpe_config.kernel_player_media,"Max number of kernel media files","INT"},
 		{ "audio-buffer-length",0,0,	G_OPTION_ARG_INT,&rtpe_config.audio_buffer_length,"Length in milliseconds of audio buffer","INT"},
 		{ "audio-buffer-delay",0,0,	G_OPTION_ARG_INT,&rtpe_config.audio_buffer_delay,"Initial delay in milliseconds for buffered audio","INT"},
 		{ "audio-player",0,0,	G_OPTION_ARG_STRING,	&use_audio_player,	"When to enable the internal audio player","on-demand|play-media|transcoding|always"},
@@ -625,6 +672,14 @@ static void options(int *argc, char ***argv) {
 		{ "socket-cpu-affinity",0,0,G_OPTION_ARG_INT,	&rtpe_config.cpu_affinity,"CPU affinity for media sockets","INT"},
 #endif
 		{ "janus-secret", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.janus_secret,"Admin secret for Janus protocol","STRING"},
+		{ "rtcp-interval", 0,0,	G_OPTION_ARG_INT,	&rtpe_config.rtcp_interval,"Delay in milliseconds between RTCP packets when generate-rtcp flag is on, where random dispersion < 1 sec is added on top","INT"},
+		{ "max-recv-iters", 0, 0, G_OPTION_ARG_INT,    &rtpe_config.max_recv_iters,  "Maximum continuous reading cycles in UDP poller loop.", "INT"},
+		{ "vsc-start-rec",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.vsc_start_rec.s,"DTMF VSC to start recording.", "STRING"},
+		{ "vsc-stop-rec",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.vsc_stop_rec.s,"DTMF VSC to stop recording.", "STRING"},
+		{ "vsc-start-stop-rec",0,0,G_OPTION_ARG_STRING,	&rtpe_config.vsc_start_stop_rec.s,"DTMF VSC to start/stop recording.", "STRING"},
+		{ "vsc-pause-rec",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.vsc_pause_rec.s,"DTMF VSC to pause recording.", "STRING"},
+		{ "vsc-pause-resume-rec",0,0,G_OPTION_ARG_STRING,&rtpe_config.vsc_pause_resume_rec.s,"DTMF VSC to pause/resume recording.", "STRING"},
+		{ "vsc-start-pause-resume-rec",0,0,G_OPTION_ARG_STRING,&rtpe_config.vsc_start_pause_resume_rec.s,"DTMF VSC to start/pause/resume recording.", "STRING"},
 
 		{ NULL, }
 	};
@@ -642,11 +697,60 @@ static void options(int *argc, char ***argv) {
 	if (rtpe_config.dtls_ciphers == NULL)
 		rtpe_config.dtls_ciphers = g_strdup("DEFAULT:!NULL:!aNULL:!SHA256:!SHA384:!aECDH:!AESGCM+AES256:!aPSK");
 
+#ifndef WITHOUT_NFTABLES
+	if (rtpe_config.nftables_chain == NULL)
+		rtpe_config.nftables_chain = g_strdup("rtpengine");
+
+	if (rtpe_config.nftables_base_chain == NULL)
+		rtpe_config.nftables_base_chain = g_strdup("INPUT");
+
+	if (!nftables_family
+			|| !strcmp(nftables_family, "ip,ip6") || !strcmp(nftables_family, "ip4,ip6")
+			|| !strcmp(nftables_family, "ip6,ip") || !strcmp(nftables_family, "ip6,ip4"))
+		rtpe_config.nftables_family = 0; // default
+	else if (!strcmp(nftables_family, "ip") || !strcmp(nftables_family, "ip4"))
+		rtpe_config.nftables_family = NFPROTO_IPV4;
+	else if (!strcmp(nftables_family, "ip6"))
+		rtpe_config.nftables_family = NFPROTO_IPV6;
+	else
+		die("Invalid value for 'nftables-family' ('%s')", nftables_family);
+#endif
+
 	if (codecs) {
 		codeclib_init(1);
 		exit(0);
 	}
 
+#ifndef WITHOUT_NFTABLES
+	int nftables_actions = nftables_start + nftables_stop + nftables_status;
+	if (nftables_actions) {
+		if (!rtpe_config.nftables_chain || !rtpe_config.nftables_chain[0])
+			die("Cannot do nftables setup without knowing which nftables chain (--nftables-chain=...)");
+		if (rtpe_config.kernel_table < 0)
+			die("Cannot do nftables setup without configured kernel table number");
+		if (nftables_actions > 1)
+			die("Cannot do more than one of --nftables-start, --nftables-stop or --nftables-status");
+		const char *err;
+		if (nftables_status) {
+			int xv = nftables_check(rtpe_config.nftables_chain, rtpe_config.nftables_base_chain,
+					(nftables_args){.family = rtpe_config.nftables_family});
+			exit(xv);
+		}
+		if (nftables_start)
+			err = nftables_setup(rtpe_config.nftables_chain, rtpe_config.nftables_base_chain,
+					(nftables_args) {.table = rtpe_config.kernel_table,
+					.append = rtpe_config.nftables_append,
+					.family = rtpe_config.nftables_family});
+		else // nftables_stop
+			err = nftables_shutdown(rtpe_config.nftables_chain, rtpe_config.nftables_base_chain,
+					(nftables_args){.family = rtpe_config.nftables_family});
+		if (err)
+			die("Failed to perform nftables action: %s (%s)", err, strerror(errno));
+		printf("Success\n");
+		exit(0);
+	}
+#endif
+
 	if (!if_a)
 		die("Missing option --interface");
 
@@ -715,12 +819,18 @@ static void options(int *argc, char ***argv) {
 			die("Invalid protocol '%s' (--homer-protocol)", homerproto);
 	}
 
+	if (rtpe_config.homer_ng_capt_proto <0 || rtpe_config.homer_ng_capt_proto > 255)
+		die("Invalid homer-ng-capture-proto value");
+
 	if (rtpe_config.default_tos < 0 || rtpe_config.default_tos > 255)
 		die("Invalid TOS value");
 
 	if (rtpe_config.control_tos < 0 || rtpe_config.control_tos > 255)
 		die("Invalid control-ng TOS value");
 
+	if (rtpe_config.max_recv_iters < 1)
+		die("Invalid max-recv-iters value");
+
 	if (rtpe_config.timeout <= 0)
 		rtpe_config.timeout = 60;
 
@@ -733,6 +843,9 @@ static void options(int *argc, char ***argv) {
 	if (rtpe_config.final_timeout <= 0)
 		rtpe_config.final_timeout = 0;
 
+	if (rtpe_config.rtcp_interval <= 0)
+		rtpe_config.rtcp_interval = 5000;
+
 	if (redisps)
 		if (redis_ep_parse(&rtpe_config.redis_ep, &rtpe_config.redis_db, &rtpe_config.redis_auth, "RTPENGINE_REDIS_AUTH_PW", redisps))
 			die("Invalid Redis endpoint [IP:PORT/INT] '%s' (--redis)", redisps);
@@ -936,25 +1049,45 @@ static void options(int *argc, char ***argv) {
 			die("Invalid --control-pmtu option ('%s')", control_pmtu);
 	}
 
-	rwlock_unlock_w(&rtpe_config.config_lock);
+#define STR_LEN_INIT(x) if (rtpe_config.x.s) rtpe_config.x.len = strlen(rtpe_config.x.s)
+	STR_LEN_INIT(vsc_start_rec);
+	STR_LEN_INIT(vsc_stop_rec);
+	STR_LEN_INIT(vsc_start_stop_rec);
+	STR_LEN_INIT(vsc_pause_rec);
+	STR_LEN_INIT(vsc_pause_resume_rec);
+	STR_LEN_INIT(vsc_start_pause_resume_rec);
+#undef STR_LEN_INIT
+
+	if (rtpe_config.num_threads < 1)
+		rtpe_config.num_threads = num_cpu_cores(4);
+	if (rtpe_config.media_num_threads < 0)
+		rtpe_config.media_num_threads = rtpe_config.num_threads;
+	if (rtpe_config.max_sessions < -1)
+		rtpe_config.max_sessions = -1;
+	if (rtpe_config.redis_num_threads < 1)
+		rtpe_config.redis_num_threads = num_cpu_cores(REDIS_RESTORE_NUM_THREADS);
+
+	if (rtpe_config.cpu_affinity < 0) {
+		rtpe_config.cpu_affinity = num_cpu_cores(0);
+		if (rtpe_config.cpu_affinity <= 0)
+			die("Number of CPU cores is unknown, cannot auto-set socket CPU affinity");
+	}
 }
 
 static void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 
-	GList* l;
 	struct intf_config* gptr_data;
 
-	rwlock_lock_w(&rtpe_config.config_lock);
-
-	for(l = rtpe_config.interfaces.head; l ; l=l->next) {
+	for (__auto_type l = rtpe_config.interfaces.head; l ; l=l->next) {
 		gptr_data = g_slice_alloc0(sizeof(*gptr_data));
-		memcpy(gptr_data, (struct intf_config*)(l->data), sizeof(*gptr_data));
-		str_init_dup(&gptr_data->name, ((struct intf_config*)(l->data))->name.s);
+		memcpy(gptr_data, l->data, sizeof(*gptr_data));
+		str_init_dup_str(&gptr_data->name, &l->data->name);
+		str_init_dup_str(&gptr_data->alias, &l->data->alias);
 
-		g_queue_push_tail(&ini_rtpe_cfg->interfaces, gptr_data);
+		t_queue_push_tail(&ini_rtpe_cfg->interfaces, gptr_data);
 	}
 
-	for(l = rtpe_config.redis_subscribed_keyspaces.head; l ; l = l->next) {
+	for (__auto_type l = rtpe_config.redis_subscribed_keyspaces.head; l ; l = l->next) {
 		// l->data has been assigned to a variable before being given into the queue structure not to get a shallow copy
 		unsigned int num = GPOINTER_TO_UINT(l->data);
 		g_queue_push_tail(&ini_rtpe_cfg->redis_subscribed_keyspaces, GINT_TO_POINTER(num));
@@ -978,6 +1111,7 @@ static void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 	ini_rtpe_cfg->redis_num_threads = rtpe_config.redis_num_threads;
 	ini_rtpe_cfg->homer_protocol = rtpe_config.homer_protocol;
 	ini_rtpe_cfg->homer_id = rtpe_config.homer_id;
+	ini_rtpe_cfg->homer_ng_capt_proto = rtpe_config.homer_ng_capt_proto;
 	ini_rtpe_cfg->no_fallback = rtpe_config.no_fallback;
 	ini_rtpe_cfg->port_min = rtpe_config.port_min;
 	ini_rtpe_cfg->port_max = rtpe_config.port_max;
@@ -1017,22 +1151,20 @@ static void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 
 	ini_rtpe_cfg->jb_length = rtpe_config.jb_length;
 	ini_rtpe_cfg->jb_clock_drift = rtpe_config.jb_clock_drift;
+	ini_rtpe_cfg->rtcp_interval = rtpe_config.rtcp_interval;
 
-	rwlock_unlock_w(&rtpe_config.config_lock);
+	ini_rtpe_cfg->max_recv_iters = rtpe_config.max_recv_iters;
 }
 
-static void
-free_config_interfaces (gpointer data)
-{
-	struct intf_config* gptr_data = data;
-
-	str_free_dup(&gptr_data->name);
-	g_slice_free1(sizeof(*gptr_data), gptr_data);
+static void free_config_interfaces(struct intf_config *i) {
+	str_free_dup(&i->name);
+	str_free_dup(&i->alias);
+	g_slice_free1(sizeof(*i), i);
 }
 
 static void unfill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 	// clear queues
-	g_queue_clear_full(&ini_rtpe_cfg->interfaces, (GDestroyNotify)free_config_interfaces);
+	t_queue_clear_full(&ini_rtpe_cfg->interfaces, free_config_interfaces);
 	g_queue_clear(&ini_rtpe_cfg->redis_subscribed_keyspaces);
 
 	// free g_strdup
@@ -1047,7 +1179,7 @@ static void unfill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 
 static void options_free(void) {
 	// clear queues
-	g_queue_clear_full(&rtpe_config.interfaces, (GDestroyNotify)free_config_interfaces);
+	t_queue_clear_full(&rtpe_config.interfaces, free_config_interfaces);
 	g_queue_clear(&rtpe_config.redis_subscribed_keyspaces);
 
 	// free config options
@@ -1088,7 +1220,62 @@ static void early_init(void) {
 	socket_init(); // needed for socktype_udp
 }
 
+#ifdef WITH_TRANSCODING
+static void clib_init(void) {
+	media_bufferpool = bufferpool_new(g_malloc, g_free, 64 * 65536);
+#ifdef HAVE_LIBURING
+	if (rtpe_config.common.io_uring)
+		uring_thread_init();
+#endif
+}
+static void clib_cleanup(void) {
+	bufferpool_destroy(media_bufferpool);
+#ifdef HAVE_LIBURING
+	if (rtpe_config.common.io_uring)
+		uring_thread_cleanup();
+#endif
+}
+static void clib_loop(void) {
+	uring_thread_loop();
+	append_thread_lpr_to_glob_lpr();
+}
+#endif
+
+static void kernel_setup(void) {
+	if (rtpe_config.kernel_table < 0)
+		goto fallback;
+#ifndef WITHOUT_NFTABLES
+	const char *err = nftables_setup(rtpe_config.nftables_chain, rtpe_config.nftables_base_chain,
+			(nftables_args) {.table = rtpe_config.kernel_table,
+			.append = rtpe_config.nftables_append,
+			.family = rtpe_config.nftables_family});
+	if (err) {
+		if (rtpe_config.no_fallback)
+			die("Failed to create nftables chains or rules: %s (%s)", err, strerror(errno));
+		ilog(LOG_ERR, "FAILED TO CREATE NFTABLES CHAINS OR RULES, KERNEL FORWARDING POSSIBLY WON'T WORK: "
+				"%s (%s)", err, strerror(errno));
+	}
+#endif
+	if (!kernel_setup_table(rtpe_config.kernel_table)) {
+		if (rtpe_config.no_fallback)
+			die("Userspace fallback disallowed - exiting");
+		goto fallback;
+	}
+
+	if (rtpe_config.player_cache && rtpe_config.kernel_player > 0 && rtpe_config.kernel_player_media > 0) {
+	       if (!kernel_init_player(rtpe_config.kernel_player_media, rtpe_config.kernel_player))
+		       die("Failed to initialise kernel media player");
+	}
+
+	return;
+
+fallback:
+	shm_bufferpool = bufferpool_new(g_malloc, g_free, 4096); // fallback userspace bufferpool
+}
+
+
 static void init_everything(void) {
+	bufferpool_init();
 	gettimeofday(&rtpe_now, NULL);
 	log_init(rtpe_common_config_ptr->log_name);
 	log_format(rtpe_config.log_format);
@@ -1106,12 +1293,18 @@ static void init_everything(void) {
 	dtls_init();
 	ice_init();
 	crypto_init_main();
+	kernel_setup();
 	interfaces_init(&rtpe_config.interfaces);
 	iptables_init();
 	control_ng_init();
 	if (call_interfaces_init())
 		abort();
 	statistics_init();
+#ifdef WITH_TRANSCODING
+	codeclib_thread_init = clib_init;
+	codeclib_thread_cleanup = clib_cleanup;
+	codeclib_thread_loop = clib_loop;
+#endif
 	codeclib_init(0);
 	media_player_init();
 	if (!dtmf_init())
@@ -1122,43 +1315,54 @@ static void init_everything(void) {
 		abort();
 	codecs_init();
 	janus_init();
+	if (!kernel_init_table())
+		die("Kernel module version mismatch or other fatal error");
 }
 
-
 static void create_everything(void) {
 	struct timeval tmp_tv;
 
 	gettimeofday(&rtpe_now, NULL);
 
-	if (rtpe_config.kernel_table < 0)
-		goto no_kernel;
-	if (kernel_setup_table(rtpe_config.kernel_table)) {
-		if (rtpe_config.no_fallback) {
-			ilog(LOG_CRIT, "Userspace fallback disallowed - exiting");
-			exit(-1);
-		}
-		goto no_kernel;
+
+	// either one global poller, or one per thread for media sockets plus one for control sockets
+#ifdef HAVE_LIBURING
+	if (rtpe_config.common.io_uring) {
+		rtpe_config.poller_per_thread = true;
+		rtpe_poller_add_item = uring_poller_add_item;
+		rtpe_poller_del_item = uring_poller_del_item;
+		rtpe_poller_del_item_callback = uring_poller_del_item_callback;
+		rtpe_poller_blocked = uring_poller_blocked;
+		rtpe_poller_isblocked = uring_poller_isblocked;
+		rtpe_poller_error = uring_poller_error;
 	}
+#endif
 
-no_kernel:
-	rtpe_poller = poller_new();
-	if (!rtpe_poller)
-		die("poller creation failed");
-
-	rtpe_poller_map = poller_map_new();
-	if (!rtpe_poller_map)
-		die("poller map creation failed");
+	if (!rtpe_config.poller_per_thread) {
+		num_media_pollers = num_rtpe_pollers = 1;
+		num_poller_threads = rtpe_config.num_threads;
+	}
+	else {
+		num_media_pollers = rtpe_config.num_threads;
+		num_rtpe_pollers = num_media_pollers + 1;
+		num_poller_threads = num_rtpe_pollers;
+	}
+	rtpe_pollers = g_malloc(sizeof(*rtpe_pollers) * num_rtpe_pollers);
+	for (unsigned int i = 0; i < num_rtpe_pollers; i++) {
+		rtpe_pollers[i] = 
+#ifdef HAVE_LIBURING
+			rtpe_config.common.io_uring ? uring_poller_new() :
+#endif
+			poller_new();
+		if (!rtpe_pollers[i])
+			die("poller creation failed");
+	}
+	rtpe_control_poller = rtpe_pollers[num_rtpe_pollers - 1];
 
 	if (call_init())
 		abort();
 
-        rwlock_init(&rtpe_config.config_lock);
-	if (rtpe_config.max_sessions < -1) {
-		rtpe_config.max_sessions = -1;
-	}
-
-	if (rtpe_config.redis_num_threads < 1)
-		rtpe_config.redis_num_threads = num_cpu_cores(REDIS_RESTORE_NUM_THREADS);
+        rwlock_init(&rtpe_config.keyspaces_lock);
 
 	create_listeners(&rtpe_config.tcp_listen_ep,     &rtpe_tcp,            (void *(*)(const endpoint_t *)) control_tcp_new,    false, "TCP control");
 	create_listeners(&rtpe_config.udp_listen_ep,     &rtpe_udp,            (void *(*)(const endpoint_t *)) control_udp_new,    true,  "UDP control");
@@ -1194,15 +1398,6 @@ no_kernel:
 			rtpe_redis_write = rtpe_redis;
 	}
 
-	if (rtpe_config.num_threads < 1)
-		rtpe_config.num_threads = num_cpu_cores(4);
-
-	if (rtpe_config.cpu_affinity < 0) {
-		rtpe_config.cpu_affinity = num_cpu_cores(0);
-		if (rtpe_config.cpu_affinity <= 0)
-			die("Number of CPU cores is unknown, cannot auto-set socket CPU affinity");
-	}
-
 	if (websocket_init())
 		die("Failed to init websocket listener");
 
@@ -1212,6 +1407,7 @@ no_kernel:
 	homer_sender_init(&rtpe_config.homer_ep, rtpe_config.homer_protocol, rtpe_config.homer_id);
 
 	rtcp_init(); // must come after Homer init
+	init_ng_tracing(); // must come after Homer init
 
 	gettimeofday(&rtpe_latest_graphite_interval_start, NULL);
 
@@ -1262,9 +1458,32 @@ static void do_redis_restore(void) {
 }
 
 
-int main(int argc, char **argv) {
-	int idx;
+#ifdef HAVE_LIBURING
+static void uring_thread_waker(struct thread_waker *wk) {
+	struct poller *p = wk->arg;
+	uring_poller_wake(p);
+}
+static void uring_poller_loop(void *ptr) {
+	struct poller *p = ptr;
 
+	uring_poller_add_waker(p);
+
+	struct thread_waker wk = {.func = uring_thread_waker, .arg = p};
+	thread_waker_add_generic(&wk);
+
+	while (!rtpe_shutdown) {
+		gettimeofday(&rtpe_now, NULL);
+		uring_poller_poll(p);
+		append_thread_lpr_to_glob_lpr();
+		log_info_reset();
+	}
+	thread_waker_del(&wk);
+	uring_poller_clear(p);
+}
+#endif
+
+
+int main(int argc, char **argv) {
 	early_init();
 	options(&argc, &argv);
 	init_everything();
@@ -1287,10 +1506,6 @@ int main(int argc, char **argv) {
 	thread_create_looper(call_rate_stats_updater, rtpe_config.idle_scheduling,
 			rtpe_config.idle_priority, "call stats", 1000000);
 
-	/* separate thread for ports iterations (stats update from the kernel) functionality */
-	thread_create_looper(kernel_stats_updater, rtpe_config.idle_scheduling,
-			rtpe_config.idle_priority, "kernel stats", 1000000);
-
 	/* separate thread for ice slow timer functionality */
 	thread_create_looper(ice_slow_timer, rtpe_config.idle_scheduling,
 			rtpe_config.idle_priority, "ICE slow", 1000000);
@@ -1318,38 +1533,26 @@ int main(int argc, char **argv) {
 		thread_create_detach(mqtt_loop, NULL, "mqtt");
 #endif
 
-	thread_create_detach(ice_thread_run, NULL, "ICE");
+	ice_thread_launch();
 
 	websocket_start();
 
 	service_notify("READY=1\n");
 
-	for (idx = 0; idx < rtpe_config.num_threads; ++idx) {
-		if (!rtpe_config.poller_per_thread)
-			thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
-		else
-			thread_create_detach_prio(poller_loop, rtpe_poller_map, rtpe_config.scheduling, rtpe_config.priority, "poller");
-	}
-
-	if (rtpe_config.poller_per_thread)
-		thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
-
-	if (rtpe_config.media_num_threads < 0)
-		rtpe_config.media_num_threads = rtpe_config.num_threads;
-	for (idx = 0; idx < rtpe_config.media_num_threads; ++idx) {
-#ifdef WITH_TRANSCODING
-		thread_create_detach_prio(media_player_loop, NULL, rtpe_config.scheduling,
-				rtpe_config.priority, "media player");
+	for (unsigned int idx = 0; idx < num_poller_threads; ++idx)
+		thread_create_detach_prio(
+#ifdef HAVE_LIBURING
+				rtpe_config.common.io_uring ? uring_poller_loop :
 #endif
-		thread_create_detach_prio(send_timer_loop, NULL, rtpe_config.scheduling,
-				rtpe_config.priority, "send timer");
-		if (rtpe_config.jb_length > 0)
-			thread_create_detach_prio(jitter_buffer_loop, NULL, rtpe_config.scheduling,
-					rtpe_config.priority, "jitter buffer");
-		thread_create_detach_prio(codec_timers_loop, NULL, rtpe_config.scheduling,
-				rtpe_config.priority, "codec timer");
-	}
+				poller_loop,
+				rtpe_pollers[idx % num_rtpe_pollers],
+				rtpe_config.scheduling, rtpe_config.priority,
+				idx < rtpe_config.num_threads ? "poller" : "cpoller");
 
+	media_player_launch();
+	send_timer_launch();
+	jitter_buffer_launch();
+	codec_timers_launch();
 
 	// reap threads as they shut down during run time
 	threads_join_all(false);
@@ -1410,9 +1613,22 @@ int main(int argc, char **argv) {
 	release_listeners(&rtpe_tcp);
 	release_listeners(&rtpe_control_ng);
 	release_listeners(&rtpe_control_ng_tcp);
-	poller_free(&rtpe_poller);
-	poller_map_free(&rtpe_poller_map);
+	for (unsigned int idx = 0; idx < num_rtpe_pollers; ++idx)
+#ifdef HAVE_LIBURING
+		if (rtpe_config.common.io_uring)
+			uring_poller_free(&rtpe_pollers[idx]);
+		else
+#endif
+			poller_free(&rtpe_pollers[idx]);
+	g_free(rtpe_pollers);
 	interfaces_free();
+#ifndef WITHOUT_NFTABLES
+	nftables_shutdown(rtpe_config.nftables_chain, rtpe_config.nftables_base_chain,
+			(nftables_args){.family = rtpe_config.nftables_family});
+#endif
+	bufferpool_destroy(shm_bufferpool);
+	kernel_shutdown_table();
+	bufferpool_cleanup();
 
 	return 0;
 }
