@@ -1,8 +1,6 @@
 #ifndef _DTLS_H_
 #define _DTLS_H_
 
-
-
 #include <time.h>
 #include <openssl/x509.h>
 #include <openssl/ssl.h>
@@ -12,20 +10,12 @@
 #include "str.h"
 #include "obj.h"
 #include "socket.h"
-
-
-
+#include "types.h"
 
 #define DTLS_MAX_DIGEST_LEN 64
 
-
-
-
 struct packet_stream;
 struct sockaddr_in6;
-struct stream_fd;
-
-
 
 struct dtls_hash_func {
 	const char *name;
@@ -70,7 +60,7 @@ struct dtls_cert *dtls_cert(void);
 void dtls_cert_free(void);
 
 int dtls_connection_init(struct dtls_connection *, struct packet_stream *, int active, struct dtls_cert *cert);
-int dtls(struct stream_fd *, const str *s, const endpoint_t *sin);
+int dtls(stream_fd *, const str *s, const endpoint_t *sin);
 void dtls_connection_cleanup(struct dtls_connection *);
 void dtls_shutdown(struct packet_stream *ps);
 
@@ -116,7 +106,7 @@ INLINE int dtls_is_active(const struct dtls_connection *d) {
 }
 
 
-struct dtls_connection *dtls_ptr(struct stream_fd *sfd);
+struct dtls_connection *dtls_ptr(stream_fd *sfd);
 
 
 
